@@ -12,8 +12,7 @@ import graphdot.cuda
 from graphdot import cpp
 from graphdot.codegen import Template
 from graphdot.codegen.cpptool import decltype
-from graphdot.cuda.array import umempty, umzeros, umarray, umlike
-from graphdot.cuda.resizable_array import ResizableArray
+from graphdot.cuda.array import umempty, umzeros, umarray
 from graphdot.microkernel import TensorProduct, Product
 from ._backend import Backend
 from ._scratch import BlockScratch
@@ -52,8 +51,6 @@ class CUDABackend(Backend):
         self.device = self.ctx.get_device()
         self.scratch = None
         self.scratch_capacity = 0
-        self.graph_cache = {}
-        self.graph_cpp = ResizableArray(OctileGraph.dtype, allocator='numpy')
 
         self.block_per_sm = kwargs.pop('block_per_sm', 8)
         self.block_size = kwargs.pop('block_size', 128)
@@ -87,20 +84,12 @@ class CUDABackend(Backend):
             self.scratch_capacity = self.scratch[0].capacity
             self.ctx.synchronize()
 
-    def _register_graph(self, graph, p):
-        if not hasattr(graph, 'cache_tag'):
-            graph.cache_tag = uuid.uuid4()
-        key = uuid.uuid5(self.uuid, graph.cache_tag.hex)
-        if key not in self.graph_cache:
+    def _register_graph(self, graph):
+        if self.uuid not in graph.cookie:
             # convert to GPU format
             og = OctileGraph(graph)
-            # assign starting probabilities
-            ps = umarray(np.array([p(*r) for r in graph.nodes.iterrows()],
-                                  dtype=np.float32))
-            i = len(self.graph_cpp)
-            self.graph_cpp.append(og.state)
-            self.graph_cache[key] = (i, ps, og)
-        return self.graph_cache[key]
+            graph.cookie[self.uuid] = (og, og.state)
+        return graph.cookie[self.uuid]
 
     def _compile(self, src):
         with warnings.catch_warnings(record=True) as w:
@@ -141,9 +130,9 @@ class CUDABackend(Backend):
 
     @staticmethod
     def gencode_kernel(kernel, name):
-        fun, jac = kernel.gen_expr('x1', 'x2', jac=True)
+        fun, jac = kernel.gen_expr('x1', 'x2')
 
-        kernel_src = Template(r'''
+        return Template(r'''
         using ${name}_theta_t = ${theta_t};
 
         struct ${name}_t : ${name}_theta_t {
@@ -170,7 +159,36 @@ class CUDABackend(Backend):
             jac=[f'j[{i}] = {expr}' for i, expr in enumerate(jac)],
         )
 
-        return kernel_src
+    @staticmethod
+    def gencode_probability(pfunc, name):
+        fun, jac = pfunc.gen_expr()
+
+        return Template(r'''
+        using ${name}_theta_t = ${theta_t};
+
+        struct ${name}_t : ${name}_theta_t {
+
+            constexpr static int jac_dims = ${jac_dims};
+
+            template<class N> __device__ __inline__
+            auto operator() (N const &n) const {
+                return ${expr};
+            }
+
+            template<class N> __device__ __inline__
+            void _j_a_c_o_b_i_a_n_(float j[], N const &n) const {
+                ${jac;\n};
+            }
+        };
+
+        __constant__ ${name}_t ${name};
+        ''').render(
+            name=name,
+            jac_dims=len(jac),
+            theta_t=decltype(pfunc),
+            expr=fun,
+            jac=[f'j[{i}] = {expr}' for i, expr in enumerate(jac)],
+        )
 
     def __call__(self, graphs, node_kernel, edge_kernel, p, q, jobs, starts,
                  output, output_shape, traits, timer):
@@ -178,18 +196,13 @@ class CUDABackend(Backend):
         timer.tic('transferring graphs to GPU')
 
         og_last = None
-        og_indices = np.empty(len(graphs), np.uint32)
-        starting_p = umempty(len(graphs), np.uintp)
-
+        graphs_d = umempty(len(graphs), dtype=OctileGraph.dtype)
         for i, g in enumerate(graphs):
-            idx, ps, og = self._register_graph(g, p)
+            og, ogstate = self._register_graph(g)
             if i > 0:
                 self._assert_homogeneous(og_last, og)
             og_last = og
-            og_indices[i] = idx
-            starting_p[i] = int(ps.base)
-
-        og_indices_d = umlike(self.graph_cpp[og_indices])
+            graphs_d[i] = ogstate
 
         weighted = og_last.weighted
         node_t = og_last.node_t
@@ -210,11 +223,13 @@ class CUDABackend(Backend):
 
         node_kernel_src = self.gencode_kernel(node_kernel, 'node_kernel')
         edge_kernel_src = self.gencode_kernel(edge_kernel, 'edge_kernel')
+        p_start_src = self.gencode_probability(p, 'p_start')
 
         with self.template.context(traits=traits) as template:
             self.source = template.render(
                 node_kernel=node_kernel_src,
                 edge_kernel=edge_kernel_src,
+                p_start=p_start_src,
                 node_t=decltype(node_t),
                 edge_t=decltype(edge_t)
             )
@@ -244,13 +259,15 @@ class CUDABackend(Backend):
         cuda.memcpy_htod(p_edge_kernel, np.array([edge_kernel.state],
                                                  dtype=edge_kernel.dtype))
 
+        p_p_start, _ = self.module.get_global('p_start')
+        cuda.memcpy_htod(p_p_start, np.array([p.state], dtype=p.dtype))
+
         timer.toc('calculating launch configuration')
 
         ''' GPU kernel execution '''
         timer.tic('GPU kernel execution')
         kernel(
-            og_indices_d,
-            starting_p,
+            graphs_d,
             self.scratch_d,
             jobs,
             starts,
